import typing

GeneratorReturnType = typing.TypeVar("GeneratorReturnType", covariant=True)
"""
Return type of the generator.
"""

CaptchaReturnType = typing.TypeVar("CaptchaReturnType")
"""
Return type of the Captcha object.
"""



from typing import Dict, List

sequence = tuple(t / 20.0 for t in range(21))
beziers: Dict[int, List[List[float]]] = {}


def pascal_row(n: int) -> List[int]:
    """Returns n-th row of Pascal's triangle"""
    result = [1]
    x, numerator = 1, n
    for denominator in range(1, n // 2 + 1):
        x *= numerator
        x = int(x / denominator)
        result.append(x)
        numerator -= 1
    if n & 1 == 0:
        result.extend(reversed(result[:-1]))
    else:
        result.extend(reversed(result))
    return result


def make_bezier(n: int) -> List[List[float]]:
    """Bezier curves:
    http://en.wikipedia.org/wiki/B%C3%A9zier_curve#Generalization
    """
    try:
        return beziers[n]
    except KeyError:
        combinations = pascal_row(n - 1)
        result: List[List[float]] = []
        for t in sequence:
            power = (t**i for i in range(n))
            powers = ((1 - t) ** i for i in range(n - 1, -1, -1))
            coefficient = [c * a * b for c, a, b in zip(combinations, power, powers)]
            result.append(coefficient)
        beziers[n] = result
        return result



import typing
from random import randint, random, uniform
from secrets import choice

from PIL import Image, ImageColor, ImageDraw, ImageFilter, ImageFont

from discapty.wheezylib.bezier import make_bezier


def captcha(
    drawings: typing.Sequence[typing.Callable[[Image.Image, str], Image.Image]],
    width: int = 200,
    height: int = 75,
) -> typing.Callable[[str], Image.Image]:
    def render(text_input: str) -> Image.Image:
        img = Image.new("RGB", (width, height), (255, 255, 255))
        for drawing in drawings:
            img = drawing(img, text_input)
        return img

    return render


# Drawings


def background(background_color: str = "#EEEECC"):
    color = ImageColor.getrgb(background_color)

    def render(image: Image.Image, _: str) -> Image.Image:
        ImageDraw.Draw(image).rectangle(((0, 0), image.size), fill=color)
        return image

    return render


def apply_filter(img_filter: ImageFilter.Filter):
    """
    Filter an image with the given filter.
    """

    def render(image: Image.Image, _: str) -> Image.Image:
        return image.filter(img_filter)

    return render


def apply_filters(filters: typing.Sequence[ImageFilter.Filter]):
    """
    Filter an image with the given filters.
    """

    def render(image: Image.Image, _: str) -> Image.Image:
        for image_filter in filters:
            image = image.filter(image_filter)
        return image

    return render


def smooth():
    def render(image: Image.Image, _: str) -> Image.Image:
        return image.filter(ImageFilter.SMOOTH)

    return render


def curve(
    curve_color: str = "#5C87B2", *, curve_width: int = 4, curve_number: int = 6
) -> typing.Callable[[Image.Image, str], Image.Image]:
    color = ImageColor.getrgb(curve_color)

    def render(image: Image.Image, _: str):
        width, height = image.size
        dx = width / curve_number
        path = [(dx * i, randint(0, height)) for i in range(1, curve_number)]
        all_coefs = make_bezier(curve_number - 1)
        points = [
            tuple(sum(coef * p for coef, p in zip(coefs, ps)) for ps in zip(*path))
            for coefs in all_coefs
        ]

        ImageDraw.Draw(image).line(points, fill=color, width=curve_width)  # type: ignore
        return image

    return render


def noise(
    noise_color: str = "#EEEECC",
    *,
    noise_number: int = 50,
    noise_width: int = 2,
) -> typing.Callable[[Image.Image, str], Image.Image]:
    color = ImageColor.getrgb(noise_color)

    def render(image: Image.Image, _: str):
        width, height = image.size
        dx = width / 10
        width = width - dx
        dy = height / 10
        height = height - dy
        draw = ImageDraw.Draw(image)

        for __ in range(noise_number):
            x = int(uniform(dx, width))
            y = int(uniform(dy, height))
            draw.line(((x, y), (x + noise_width, y)), fill=color, width=noise_width)

        return image

    return render


def text(
    fonts: typing.Sequence[str],
    fonts_sizes: typing.Tuple[int, ...],
    *,
    drawings: typing.Optional[typing.Sequence[typing.Callable[[Image.Image], Image.Image]]],
    text_color: str = "#5C87B2",
    squeeze_factor: float = 0.8,
) -> typing.Callable[[Image.Image, str], Image.Image]:
    tt_fonts = tuple(ImageFont.truetype(name, size) for name in fonts for size in fonts_sizes)

    color = ImageColor.getrgb(text_color)

    def render(image: Image.Image, text_input: str) -> Image.Image:
        draw = ImageDraw.Draw(image)

        # We will make a loop for each character in input, so we can
        # use drawings on each character and create a more wheezy-like
        # result.
        # In the end, we will paste those characters in the image.
        characters: typing.List[Image.Image] = []

        for input_character in text_input:
            # Mostly setting up things here...
            rand_font = choice(tt_fonts)
            _, _, char_width, char_height = draw.textbbox((0, 0), input_character, font=rand_font)
            character = Image.new("RGB", (char_width, char_height), (0, 0, 0))
            character_draw = ImageDraw.Draw(character)

            # Draw the actual character
            character_draw.text((0, 0), input_character, fill=color, font=rand_font)  # type: ignore
            character = character.crop(character.getbbox())

            # Applies drawings
            if drawings:
                for drawing in drawings:
                    character = drawing(character)

            characters.append(character)

        width, height = image.size
        image_offset = int(
            (
                width
                - sum(int(i.size[0] * squeeze_factor) for i in characters[:-1])
                - characters[-1].size[0]
            )
            / 2
        )

        # Paste characters into final image
        for character in characters:
            character_width, character_height = character.size
            mask = character.convert("L").point(lambda x: int(x * 1.97))  # type: ignore
            image.paste(character, (image_offset, int((height - character_height) / 2)), mask)
            image_offset += int(character_width * squeeze_factor)

        return image

    return render


def warp(
    dx_factor: float = 0.27, dy_factor: float = 0.21
) -> typing.Callable[[Image.Image], Image.Image]:
    def render(image: Image.Image) -> Image.Image:
        width, height = image.size
        dx = width * dx_factor
        dy = height * dy_factor
        x1 = int(uniform(-dx, dx))
        y1 = int(uniform(-dy, dy))
        x2 = int(uniform(-dx, dx))
        y2 = int(uniform(-dy, dy))

        new_image = Image.new("RGB", (width + abs(x1) + abs(x2), height + abs(y1) + abs(y2)))
        new_image.paste(image, (abs(x1), abs(y1)))
        width, height = new_image.size
        return new_image.transform(
            (width, height),
            Image.Transform.QUAD,
            (
                x1,
                y1,
                -x1,
                width - y2,
                height + x2,
                width + y2,
                height - x2,
                -y1,
            ),
        )

    return render


def offset(
    dx_factor: float = 0.1, dy_factor: float = 0.2
) -> typing.Callable[[Image.Image], Image.Image]:
    def render(image: Image.Image) -> Image.Image:
        width, height = image.size
        dx = int(random() * width * dx_factor)
        dy = int(random() * height * dy_factor)
        new_image = Image.new("RGB", (width + dx, height + dy))
        new_image.paste(image, (dx, dy))
        return new_image

    return render


def rotate(angle: int = 25) -> typing.Callable[[Image.Image], Image.Image]:
    def render(image: Image.Image):
        return image.rotate(uniform(-angle, angle), Image.Resampling.BILINEAR, expand=True)

    return render



import pathlib
import typing
from abc import ABC, abstractmethod
from os import listdir
from os.path import abspath, dirname, isfile, join
from random import choice, randint, random, uniform

import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw
import PIL.ImageFilter
import PIL.ImageFont
import pydantic
from pydantic.color import Color

from discapty.constants import GeneratorReturnType
from discapty.utils import random_color

from .wheezylib import image as wheezy_captcha

PATH: str = join(abspath(dirname(__file__)), "fonts")
DEFAULT_FONTS: typing.List[str] = [join(PATH, f) for f in listdir(PATH) if isfile(join(PATH, f))]


class Generator(ABC, pydantic.BaseModel, typing.Generic[GeneratorReturnType]):
    """
    Base class for all generators.

    A generator is used to especially generate a Captcha object based on a given text.
    A generator looks like this:

    .. code-block:: python

       class MyGenerator(Generator[str]):
           def generate(self, text: str) -> str:
               return "+".join(text)


    A generator can be supplied with parameters using class's attributes, for example:


    .. code-block:: python

       class MyGenerator(Generator[str]):
           separator = "+"

           def generate(self, text: str) -> str:
               return self.separator.join(text)

       gen1 = MyGenerator()  # Separator here is "+"
       gen2 = MyGenerator(separator="-")  # Separator here is "-"


    Here, separator has a default value, which can be overridden by the user, or not.
    If you wish to create a generator with a required value, you can use "...", like this:


    .. code-block:: python

       class MyGenerator(Generator[str]):
           separator: str = ...

           ...

       MyGenerator(separator="+")  # Works! 👍
       MyGenerator()  # Raises an error! 👎


    If you wish to know more on that subject, visit Pydantic's documentation as this is what
    :py:obj:`Generator` uses under the hood.
    https://pydantic-docs.helpmanual.io/

    .. versionadded:: 2.0.0

    .. versionchanged:: 2.1.0

       This class is now a generic class, taking as the parameter the type output of ".generate".
    """

    class Config:
        """
        :meta private:
        legit no one cares about this class
        """

        validate_all = True
        validate_assignment = True

    @property
    def required_keys(self) -> typing.List[str]:
        """
        List of all child's required keys.

        Returns
        -------
        List of :py:class:`str` :
            The list of required keys.
        """
        return [key for key, value in self.__fields__.items() if value.required]

    @property
    def optional_keys(self) -> typing.List[str]:
        """
        List of all child's optional keys.

        Returns
        -------
        List of :py:class:`str` :
            The list of optional keys.
        """
        return [key for key, value in self.__fields__.items() if not value.required]

    @abstractmethod
    def generate(self, text: str) -> GeneratorReturnType:
        """
        A method that needs to be implemented by the child class.
        This method will return the Captcha that the user has requested. See class's docstring.

        Parameters
        ----------
        text : :py:class:`str`
            The text to generate the captcha with.
        """
        raise NotImplementedError()


class WheezyGenerator(Generator[PIL.Image.Image]):
    """
    A wheezy image Captcha generator. Comes with many customizable settings.
    Easier to read than Image.

    Example: https://imgur.com/a/l9V09PN
    """

    fonts: typing.Sequence[typing.Union[pydantic.FilePath, str]] = DEFAULT_FONTS
    fonts_size: typing.Tuple[int, ...] = (50,)
    width: int = 300
    height: int = 125
    background_color: Color = "#EEEECC"  # type: ignore
    text_color: Color = "#5C87B2"  # type: ignore
    text_squeeze_factor: float = 0.8
    noise_number: int = 30
    noise_color: Color = "#EEEECC"  # type: ignore
    noise_level: int = 2

    @pydantic.validator("fonts_size")
    def as_many_size_as_fonts(
        cls, v: typing.Tuple[int, ...], values: typing.Dict[str, typing.Any]
    ):
        if "fonts" in values and len(v) != len(values["fonts"]):
            raise ValueError("The number of fonts_size must be equal to the number of fonts")
        return v

    def generate(self, text: str) -> PIL.Image.Image:
        """
        Generate a wheezy image. See https://imgur.com/a/l9V09PN.

        Parameters
        ----------
        text: :py:class:`str`
            The text to generate the captcha with.

        Returns
        -------
        :py:class:`PIL.Image.Image` :
            The captcha image.
        """
        fonts: typing.List[str] = [
            font if isinstance(font, str) else font.absolute().as_posix() for font in self.fonts
        ]

        fn = wheezy_captcha.captcha(
            drawings=[
                wheezy_captcha.background(self.background_color.as_hex()),
                wheezy_captcha.text(
                    fonts=fonts,
                    fonts_sizes=self.fonts_size,
                    drawings=[
                        wheezy_captcha.warp(),
                        wheezy_captcha.rotate(),
                        wheezy_captcha.offset(),
                    ],
                    text_color=self.text_color.as_hex(),
                    squeeze_factor=self.text_squeeze_factor,
                ),
                wheezy_captcha.curve(),
                wheezy_captcha.noise(
                    noise_number=self.noise_number,
                    noise_color=self.noise_color.as_hex(),
                    noise_width=self.noise_level,
                ),
                wheezy_captcha.smooth(),
            ],
            width=self.width,
            height=self.height,
        )
        return fn(text)


class ImageGenerator(Generator[PIL.Image.Image]):
    """
    An image Captcha generator. Comes with many customizable settings.
    More harder than the Wheezy generator.

    Example: https://imgur.com/a/wozYgW0
    """

    fonts: typing.Sequence[typing.Union[pydantic.FilePath, str]] = DEFAULT_FONTS
    fonts_size: typing.Tuple[int, ...] = (50,)

    background_color: Color = pydantic.Field(random_color(238, 255))
    text_color: Color = pydantic.Field(random_color(10, 200, 100))
    number_of_dots: int = 100
    width_of_dots: int = 3
    number_of_curves: int = 1
    width: int = 300
    height: int = 125

    def get_truefonts(self) -> typing.Tuple[PIL.ImageFont.FreeTypeFont, ...]:
        return self._fetch_truefonts(list(self.fonts), self.fonts_size)

    @staticmethod
    def _fetch_truefonts(
        fonts: typing.Sequence[typing.Union[pydantic.FilePath, str]],
        fonts_sizes: typing.Tuple[int, ...],
    ) -> typing.Tuple[PIL.ImageFont.FreeTypeFont, ...]:
        return tuple(
            PIL.ImageFont.truetype(n, s)
            for n in [f.absolute().as_posix() if isinstance(f, pathlib.Path) else f for f in fonts]
            for s in fonts_sizes
        )

    @staticmethod
    def create_noise_curve(
        image: PIL.Image.Image, color: Color, number: int = 1
    ) -> PIL.Image.Image:
        w, h = image.size
        while number:
            x1 = randint(0, int(w / 5))
            x2 = randint(w - int(w / 5), w)
            y1 = randint(int(h / 5), h - int(h / 5))
            y2 = randint(y1, h - int(h / 5))
            points = [x1, y1, x2, y2]
            end = randint(160, 200)
            start = randint(0, 20)
            PIL.ImageDraw.Draw(image).arc(
                points, start, end, fill=PIL.ImageColor.getrgb(color.as_hex())
            )
            number -= 1
        return image

    @staticmethod
    def create_noise_dots(
        image: PIL.Image.Image, color: Color, width: int = 3, number: int = 30
    ) -> PIL.Image.Image:
        draw = PIL.ImageDraw.Draw(image)
        w, h = image.size
        while number:
            x1 = randint(0, w)
            y1 = randint(0, h)
            pos = ((x1, y1), (x1 - 1, y1 - 1))
            draw.line(pos, fill=PIL.ImageColor.getrgb(color.as_hex()), width=width)
            number -= 1
        return image

    def create_captcha_image(self, *, chars: str) -> PIL.Image.Image:
        image = PIL.Image.new(
            "RGB",
            (self.width, self.height),
            PIL.ImageColor.getrgb(self.background_color.as_hex()),
        )
        draw = PIL.ImageDraw.Draw(image)

        def _draw_character(char: str) -> PIL.Image.Image:
            font = choice(self.get_truefonts())
            _, _, wid, hei = draw.textbbox((0, 0), char, font=font)

            dx = randint(0, 4)
            dy = randint(0, 6)
            im = PIL.Image.new("RGBA", (wid + dx, hei + dy))
            PIL.ImageDraw.Draw(im).text(  # type: ignore
                (dx, dy),
                char,
                font=font,
                fill=PIL.ImageColor.getrgb(self.text_color.as_hex()),
            )

            # Rotate
            im = im.crop(im.getbbox())
            im = im.rotate(uniform(-30, 30), PIL.Image.Resampling.BILINEAR, expand=True)

            # Warp
            dx = wid * uniform(0.1, 0.3)
            dy = hei * uniform(0.2, 0.3)
            x1 = int(uniform(-dx, dx))
            y1 = int(uniform(-dy, dy))
            x2 = int(uniform(-dx, dx))
            y2 = int(uniform(-dy, dy))
            w2 = wid + abs(x1) + abs(x2)
            h2 = hei + abs(y1) + abs(y2)
            data = (
                x1,
                y1,
                -x1,
                h2 - y2,
                w2 + x2,
                h2 + y2,
                w2 - x2,
                -y1,
            )
            im = im.resize((w2, h2))
            im = im.transform((wid, hei), PIL.Image.Transform.QUAD, data)
            return im

        images: typing.List[PIL.Image.Image] = []

        for c in chars:
            if random() > 0.5:
                images.append(_draw_character(" "))
            images.append(_draw_character(c))

        text_width = sum(im.size[0] for im in images)

        width = max(text_width, self.width)
        image = image.resize((width, self.height))

        average = int(text_width / len(chars))
        rand = int(0.25 * average)
        offset = int(average * 0.1)

        for im in images:
            w, h = im.size
            mask = im.convert("L").point([int(i * 1.97) for i in range(256)])  # type: ignore
            image.paste(im, (offset, int((self.height - h) / 2)), mask)
            offset = offset + w + randint(-rand, 0)

        image = image.resize((width, self.height))

        return image

    def generate(self, text: str) -> PIL.Image.Image:
        """
        Generate a Captcha image. See https://imgur.com/a/wozYgW0

        Parameters
        ----------
        text: :py:class:`str`
            The text to generate the captcha with.

        Returns
        -------
        :py:class:`PIL.Image.Image`
            The captcha image.
        """
        im = self.create_captcha_image(chars=text)
        self.create_noise_dots(im, self.text_color, self.width_of_dots, self.number_of_dots)
        self.create_noise_curve(im, self.text_color, self.number_of_curves)
        im = im.filter(PIL.ImageFilter.SMOOTH)
        return im


class TextGenerator(Generator[str]):
    """
    A text-based Captcha generator.
    Most insecure, but it is the most tricky.

    It adds a specific separator between each character of the given text.
    The default separator is an invisible space. (\\\\u200B)
    """

    separator: typing.Union[str, typing.List[str]] = "\u200B"

    def generate(self, text: str) -> str:
        """
        Generate a Captcha text.

        Parameters
        ----------
        text: :py:class:`str`
            The text to generate the captcha with.

        Returns
        -------
        :py:class:`str` :
            The captcha text.
        """
        if isinstance(self.separator, str):
            return self.separator.join(text)
        new_string: str = ""
        for position, character in enumerate(text):
            char = choice(self.separator)
            new_string += character + (char if position < len(text) - 1 else "")
        return new_string



from pathlib import Path
from random import choices, randint
from string import ascii_uppercase, digits
from typing import List, Optional, Union

from PIL.ImageColor import getrgb
from PIL.ImageFont import truetype


def check_fonts(*fonts: Union[str, Path]) -> Optional[List[str]]:
    """Check the given fonts by loading them. If any of them fails, return their path.

    Parameters
    ----------
    fonts : :py:class:`str` or :py:class:`pathlib.Path`
        The list of fonts to check.

    Returns
    -------
    Optional, list of :py:class:`str` :
        A list of path of the non-loadable font, if any.
    """
    failures: List[str] = []
    for font in fonts:
        if isinstance(font, Path):
            font = font.absolute().as_posix()
        try:
            truetype(font)
        except OSError:
            failures.append(font)
    return failures or None


def random_color(start: int = 0, end: int = 255, opacity: int = 0) -> str:
    # sourcery skip: comprehension-to-generator, invert-any-all
    # Incorrect suggestion
    """Returns a random color in hexadecimal format.

    Parameters
    ----------
    start : Optional, :py:class:`int`
        The starting number of the color, must be contained between 0 and 255, by default 0
    end : Optional, :py:class:`int`
        The ending number of the color, must be contained between 0 and 255, by default 255
    opacity : Optional, :py:class:`int`
        The color's opacity, by default 0

    Returns
    -------
    :py:class:`str` :
        The hexadecimal color.

    Raises
    ------
    :py:exc:`ValueError` :
        If any of `start`, `end` or `opacity` is not contained between 0 and 255.
    """
    if not all([0 <= attribute <= 255 for attribute in (start, end, opacity)]):
        raise ValueError("start, end and opacity parameters must be contained between 0 and 255.")

    # Simply return a random number...
    def rc() -> int:
        return randint(start, end)

    # Convert numbers to hexadecimal
    return "#%02X%02X%02X%02X" % (rc(), rc(), rc(), opacity)


def random_code(characters_length: Optional[int] = None) -> str:
    """
    Return a random code with the needed length.

    Parameters
    ----------
    characters_length : Optional, :py:class:`int`
        The length of the returned code.

    Returns
    -------
    :py:class:`str` :
        The random code.
    """
    return "".join(choices(ascii_uppercase + digits, k=characters_length or 4))


def validate_str_to_hex(color: str) -> bool:
    """
    Validate a string to be a valid hexadecimal color.

    Parameters
    ----------
    color : :py:class:`str`:
        The string to validate.

    Returns
    -------
    :py:class:`bool` :
        If the color is hexadecimal or not.
    """
    if not color:
        # Empty string
        return False
    try:
        getrgb(color)
    except ValueError:
        return False
    return True



from .captcha import Captcha as Captcha
from .captcha_queue import CaptchaQueue as CaptchaQueue
from .challenge import Challenge as Challenge
from .challenge import States as States
from .errors import *
from .generators import Generator as Generator
from .generators import ImageGenerator as ImageGenerator
from .generators import TextGenerator as TextGenerator
from .generators import WheezyGenerator as WheezyGenerator

# DisCapTy: Version 2.1.1
__version__ = "2.1.1"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['discapty', 'discapty.wheezylib']

package_data = \
{'': ['*'], 'discapty': ['fonts/*']}

install_requires = \
['pillow>=9.0.0,<10.0.0', 'pydantic>=1.9.1,<2.0.0']

extras_require = \
{'docs': ['furo>=2022.6.21,<2023.0.0',
          'Sphinx>=4.4.0,<5.0.0',
          'sphinx-copybutton>=0.5.0,<0.6.0']}

setup_kwargs = {
    'name': 'discapty',
    'version': '2.1.1',
    'description': 'DisCapTy help you generate obfuscated images, leaving your code easy.',
    'long_description': '# DisCapTy\n\n![DisCapTy\'s Logo](.github/logo.png#gh-dark-mode-only)![DisCapTy\'s Logo](.github/logo-dark.png#gh-light-mode-only)\n\nDisCapTy is a highly type hinted Python module to generate Captcha images without struggling your mind on how to make your own. Everyone can use it!\n\n**Documentation:** <https://discapty.readthedocs.io/>\n\n<div align="center">\n    <a href="https://pypi.org/project/DisCapTy/">\n        <img src="https://img.shields.io/pypi/v/discapty?style=flat-square" alt="DisCapTy\'s Version" />\n        <img src="https://img.shields.io/pypi/pyversions/discapty?style=flat-square" alt="Python Version Required" />\n        <img src="https://img.shields.io/pypi/dm/discapty?color=blue&style=flat-square" alt="DisCapTy\'s download" />\n    </a>\n    <a href="https://discapty.readthedocs.io/en/latest/?badge=latest">\n        <img src="https://readthedocs.org/projects/discapty/badge/?version=latest&style=flat-square" alt="Documentation Status" />\n    </a>\n</div>\n\n## Installing\n\nDisCapTy is available on PyPi!\n\n```sh\npip3 install discapty\n```\n\nTo use DisCapTy, you need a Python version equal or greater to `3.7` and below `3.11`.\n\n## Clone & Test the project\n\nThis project is dependant of [Poetry](https://python-poetry.org), a dependency management tool. You are most likely going to require this tool to correctly interact with the project & its dependencies, check out [Poetry\'s documentation](https://python-poetry.org/docs) for how to install it.\n\nTo clone the repository: `git clone https://github.com/Predeactor/DisCapTy.git`\n\nTo install dependencies: `poetry install`\n\nTo run tests: `poetry run python -m unittest`\n\n## Creating Captcha\n\nFor DisCapTy, a Captcha is simply a code with any possible objects that can be returned, for example, it is one code (Like "12345") with an image (Usually a `PIL.Image.Image` object)\nThis is because DisCapTy uses the concept of generators that are used to generate a captcha from a given code, and it can return anything.\n\nDisCapTy comes with 3 builtin generators:\n\n- TextGenerator : Text based captcha\n- WheezyGenerator : Image based captcha\n- ImageGenerator : Image based captcha\n\n### Creating Captcha manually\n\n```py\nimport discapty\n\ndef generate_a_captcha(initial_input: str) -> discapty.Captcha:\n    # This generator returns an obfuscated text.\n    captcha_for_user = discapty.TextGenerator().generate(initial_input)\n    # Create a Captcha object, the first argument is the clear code, the second is the obfuscated code. Anything goes.\n    return discapty.Captcha(initial_input, captcha_for_user)\n\n# Generate your Captcha.\ncaptcha = generate_a_captcha("12345")\n\n# Show the obfuscated code. See https://discapty.readthedocs.io for more information on this object.\nshow_captcha_to_user(captcha.captcha)\n```\n\n### Checking user\'s input\n\n```py\nimport discapty\n\n# Generate your Captcha.\ncaptcha: discapty.Captcha = generate_a_captcha("12345")\n\n# This is your user\'s input here\nuser_input: str = \'12345\'\n\nif captcha.check(user_input) is True:\n    # The user input is correct\n    print("Correct!")\nelse:\n    # The user input is incorrect\n    print("Incorrect!")\n```\n\nWhat\'s great with the `.check` method is that you can specify if you need to remove space in the user\'s input and/or check casing.\n\nCreating Captcha manually is not a recommended way, because DisCapTy comes with its opinionated challenge runner & is inefficient anyway.\n\n### Create a Challenge\n\n```py\nimport discapty\n\nchallenge = discapty.Challenge(discapty.TextGenerator(), retries=3)\n\ncaptcha = challenge.begin()\n\n# We cannot provide typehint here, `captcha` is a `typing.Any` and cannot help you, it\'ll be your\n# job to know what you\'ll get as a captcha.\nsend_captcha_to_user(captcha)\nuser_input: str = get_user_input()\n\nis_correct: bool = challenge.check(user_input)\n# If the user\'s input is correct, the challenge ends, if not, `challenge.attempted_tries` will get\n# +1, and if it is greater than the retries that has been set, then an error is raised when using\n# `.check`\n```\n\nPlease see the [documentation](https://discapty.readthedocs.io/) for more information on how the library work.\n\n## Contact\n\nYou can join my Discord server for any help: <https://discord.gg/aPVupKAxxP>\n\nDisCapTy is an open-source project distributed under the MIT license:\n![PyPI - License](https://img.shields.io/pypi/l/discapty?style=flat-square)\n\nDisCapTy uses the [Roboto](https://fonts.google.com/specimen/Roboto) font as default font.\n[This font](https://fonts.google.com/specimen/Roboto) is licensed under [Apache-2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'Predeactor',
    'author_email': 'predeactor0@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Predeactor/DisCapTy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import random
import typing

from discapty.challenge import Challenge
from discapty.constants import GeneratorReturnType
from discapty.errors import NonexistingChallengeError
from discapty.generators import Generator


class CaptchaQueue(typing.Generic[GeneratorReturnType]):
    """
    A safe handler for taking cares of managing the challenges for the developer.

    It basically offers a sane & internal way to manage your captcha using a key-value pair
    without ever having to touch the challenges/captcha directly.

    Parameters
    ----------
    generators : :py:class:`discapty.generators.Generator` or list of :py:class:`discapty.generators.Generator`
        A list or a single generator to use for creating the challenges.
        If a list is given, a random generator will be picked up when using :py:func:`create_challenge <discapty.captcha_queue.CaptchaQueue.create_challenge>`.

        You should be aware that inconsistency will occur this way, as if one generator can return
        a specific type and another one could return another kind of type.

    queue : Dict[:py:class:`str`, :py:class:`discapty.challenge.Challenge`]
        Import an existing queue. Shouldn't be required.

    Raises
    ------
    :py:exc:`ValueError` :
        If no generators has been passed.


    .. versionadded:: 2.0.0

    .. versionchanged:: 2.1.0

       This class is now a generic class and requires to indicate which type it will receive.
       (If necessary)
       If the type is not especially indicated in your variable, it should be automatically done.

    """

    generators: typing.List[Generator[GeneratorReturnType]]
    queue: typing.Dict[str, Challenge[GeneratorReturnType]]
    __total_challenges: int

    def __init__(
        self,
        generators: typing.Union[
            Generator[GeneratorReturnType],
            typing.Iterable[Generator[GeneratorReturnType]],
        ],
        *,
        queue: typing.Optional[typing.Dict[str, Challenge[GeneratorReturnType]]] = None,
    ) -> None:
        self.generators = []
        if isinstance(generators, Generator):
            self.generators.append(generators)
        else:
            self.generators.extend(iter(generators))
        self.queue = queue or {}
        self.__total_challenges = 0

    def create_challenge(
        self,
        challenge_id: typing.Optional[str] = None,
        *,
        retries: typing.Optional[int] = None,
        code: typing.Optional[str] = None,
        code_length: typing.Optional[int] = None,
    ) -> Challenge[GeneratorReturnType]:
        """
        Create a challenge for an id. Overwrite the challenge created before, unless the
        challenge is not fully completed.

        Parameters
        ----------
        challenge_id: Optional, :py:class:`str`
            The id associated to the challenge. If not given, a random id will be generated.
        retries: Optional, :py:class:`int`
            The number of allowed retries. Defaults to 3.
        code: Optional, :py:class:`str`
            The code to use. Defaults to a random code.
        code_length: Optional, :py:class:`int`
            The length of the code to generate if no code is supplied. Defaults to 4.

        Returns
        -------
        :py:class:`discapty.challenge.Challenge` :
            The generated challenge.


        .. versionchanged:: 2.1.0

           The return type will now be dynamically acquired and adapt to the given generator(s).
        """
        challenge_id = challenge_id or str(self.__total_challenges)

        random_generator = random.choice(self.generators)
        challenge = Challenge(
            random_generator,
            challenge_id,
            allowed_retries=retries,
            code=code,
            code_length=code_length,
        )
        self.queue[str(challenge_id)] = challenge

        self.__total_challenges += 1
        return challenge

    def get_challenge(self, challenge_id: str) -> Challenge[GeneratorReturnType]:
        """
        Get the challenge of an id, if it exist.

        Parameters
        ----------
        challenge_id : :py:class:`str`
            The id associated to the challenge.

        Raises
        ------
        :py:exc:`~errors.UnexistingChallengeError`:
            If the given id does not have any associated challenge.

        Returns
        -------
        :py:class:`discapty.challenge.Challenge` :
            The challenge associated to the id.


        .. versionchanged:: 2.1.0

           The return type will now be dynamically acquired and adapt to the given generator(s).

        """
        try:
            return self.queue[challenge_id]
        except KeyError as e:
            raise NonexistingChallengeError(
                f"Challenge with id '{challenge_id}' does not exist. Have you used an int?"
            ) from e

    def delete_challenge(self, challenge_id: str) -> None:
        """Delete a challenge of an id, if it exist.

        Parameters
        ----------
        challenge_id: :py:class:`int`
            The id associated to the challenge.

        Raises
        ------
        :py:exc:`~errors.UnexistingChallengeError`:
            If the given id does not have any associated challenge.
        """
        challenge = self.get_challenge(challenge_id)
        challenge.cancel()
        del self.queue[challenge_id]



import typing

from .constants import CaptchaReturnType


class Captcha(typing.Generic[CaptchaReturnType]):
    """
    Represent a Captcha object.


    .. versionchanged:: 2.0.0
       The Captcha object is no longer what creates the Captcha image, it just is the
       representation of the Captcha that the user will face.

    .. versionchanged:: 2.1.0
       This class is now a generic class and requires to indicate which type it will receive.
       (If necessary)
    """

    code: str
    captcha_object: CaptchaReturnType
    type: typing.Type[CaptchaReturnType]

    def __init__(
        self,
        code: str,
        captcha_object: CaptchaReturnType,
    ) -> None:
        self.code = code
        self.captcha_object = captcha_object
        self.type = type(self.captcha_object)

    def __repr__(self) -> str:
        return f"<Captcha type={self.type}>"

    def check(self, text: str, *, force_casing: bool = False, remove_spaces: bool = True) -> bool:
        """
        Check if a text is correct against the captcha code.

        Parameters
        ----------
        text : str
            The answer to check against the Captcha's code.
        force_casing : bool
            If True, the casing must be respected. Defaults to False.
        remove_spaces : bool
            If True, spaces will be removed when checking the answer. Defaults to True.

        Return
        ------
        bool:
            True if the answer is correct, False otherwise.
        """
        # Remove spaces if needed.
        text = text.replace(" ", "") if remove_spaces else text

        # Lower the text if needed
        text = text if force_casing else text.lower()
        code = self.code if force_casing else self.code.lower()

        # Result of the check.
        return code == text



class NonexistingChallengeError(KeyError):
    """
    Raised when trying to get a challenge that does not exist.
    Subclass of "KeyError" as this error will appear when trying to get the challenge from a dict.
    """


class InvalidFontError(Exception):
    """
    Raised when one or more fonts are invalid.
    """


class ChallengeCompletionError(Exception):
    """
    Raised when a challenge has an issue regarding its completion.
    """


class TooManyRetriesError(ChallengeCompletionError):
    """
    Raised when a challenge received more retries than allowed.
    """


class AlreadyCompletedError(ChallengeCompletionError):
    """
    Raised when a challenge has already been completed.
    """


class AlreadyRunningError(ChallengeCompletionError):
    """
    Raised when a challenge is already running.
    """



import uuid
from enum import Enum
from typing import Generic, Optional

from discapty.captcha import Captcha
from discapty.constants import GeneratorReturnType
from discapty.errors import (
    AlreadyCompletedError,
    AlreadyRunningError,
    ChallengeCompletionError,
    TooManyRetriesError,
)
from discapty.generators import Generator
from discapty.utils import random_code


class FailReason(Enum):
    """
    An enum with all possible reasons of failing the captcha.
    """

    TOO_MANY_RETRIES = "Too many retries"
    CANCELLED = "Challenge has been cancelled"


class States(Enum):
    """
    An enum representing the different states of a challenge.

    Available states are:

    * PENDING   : The challenge is waiting to begin.
    * WAITING   : The challenge is waiting for user's input.
    * COMPLETED : The challenge has been completed.
    * FAILED    : The challenge has been failed without trouble.
    * FAILURE   : The challenge has been completed without user's input and in an unexpected way. (e.g. manually cancelled)

    """

    PENDING = "Pending"
    WAITING = "Waiting"
    COMPLETED = "Completed"
    FAILED = "Failed"
    FAILURE = "Failure (Unexpected)"


class Challenge(Generic[GeneratorReturnType]):
    """
    Representation of a challenge. A challenge represent the user's Captcha
    question-answer he must face.

    This class takes cares of:
        - Generating the captcha
        - Verify inputs
        - Manage the "Captcha" object

    It frees your mind from managing all the process of a captcha challenge, keeping
    your code short and easy.

    Parameters
    ----------
    generator : Subclass of :py:class:`discapty.generators.Generator`
        The generator class to use.

    challenge_id : Optional, :py:class:`str`
        The id of the challenge. Can be a string or an id.
        If none is supplied, a random :py:func:`UUID <uuid.uuid4>` will be generated.

    allowed_retries : Optional, :py:class:`int`
        The number of retries allowed. Defaults to 3.

    code : Optional, :py:class:`str`
        The code to use. If none is supplied, a random code will be generated.

    code_length : Optional, :py:class:`int`
        The length of the code to generate if no code is supplied. Defaults to 4.


    .. versionadded:: 2.0.0

    .. versionchanged:: 2.1.0

       This class is now a generic class and requires to indicate which type it will receive.
       (If necessary)
       If the type is not especially indicated in your variable, it should be automatically done.
    """

    generator: Generator[GeneratorReturnType]
    """
    The generator used with this challenge.
    """
    code: str
    """
    The clear code.
    """
    challenge_id: str
    """
    The ID of this challenge.
    """
    allowed_retries: int
    """
    The total allowed retried of this challenge.
    """
    failures: int
    """
    The total failures since this challenge has been created.
    """
    attempted_tries: int
    """
    The total attempted tried since this challenge has been created.
    """
    state: States
    """
    The actual state of the challenge.
    """
    fail_reason: Optional[str]
    """
    The fail reason, if applicable.
    """
    __last_captcha_class: Optional[Captcha[GeneratorReturnType]]
    __last_code: Optional[str]

    def __init__(
        self,
        generator: Generator[GeneratorReturnType],
        challenge_id: Optional[str] = None,
        *,
        allowed_retries: Optional[int] = None,
        code: Optional[str] = None,
        code_length: Optional[int] = None,
    ) -> None:
        self.generator = generator

        self.code = code or random_code(code_length)
        self.challenge_id = str(challenge_id or uuid.uuid4())

        self.allowed_retries = allowed_retries or 3
        self.failures = 0
        self.attempted_tries = 0

        self.state = States.PENDING
        self.fail_reason = None

        self.__last_captcha_class = None
        self.__last_code = None

    def __repr__(self) -> str:
        return (
            f"<Challenge id={self.challenge_id} state={self.state} "
            "is_completed={self.is_completed}>"
        )

    def _set_state(
        self,
        state: States,
        fail_reason: Optional[FailReason] = None,
    ) -> None:
        self.state = state
        if (
            self.state
            in (
                States.FAILED,
                States.FAILURE,
            )
            and fail_reason
        ):
            self.fail_reason = fail_reason.value

    @property
    def _can_be_modified(self) -> bool:
        return self.state in (
            States.PENDING,
            States.WAITING,
        )

    def _create_captcha(self) -> Captcha[GeneratorReturnType]:
        if not self.__last_captcha_class or (self.code != self.__last_code):
            generated_captcha = self.generator.generate(self.code)
            self.__last_captcha_class = Captcha(self.code, generated_captcha)
            self.__last_code = self.code
        return self.__last_captcha_class

    @property
    def captcha_object(self) -> GeneratorReturnType:
        """
        Get the Captcha object.

        Returns
        -------
        :py:attr:`discapty.constants.GeneratorReturnType` :
            The Captcha object.
        """
        return self.captcha.captcha_object

    @property
    def captcha(self) -> Captcha[GeneratorReturnType]:
        """
        The Captcha class associated to this challenge.

        Returns
        -------
        :py:class:`discapty.captcha.Captcha` :
            The Captcha class.
        """
        return self._create_captcha()

    @property
    def is_completed(self) -> bool:
        """
        Check if the challenge has been completed or failed.

        Returns
        -------
        :py:class:`bool` :
            If the challenge has been completed or failed.
        """
        return self.state in (States.COMPLETED, States.FAILED)

    @property
    def is_correct(self) -> Optional[bool]:
        """
        Check if the challenge has been completed. If not, return None. If failed, return False.

        Returns
        -------
        Optional, :py:class:`bool` :
            If the challenge has been completed with success. If not, return False. If not completed, return None.
        """
        return self.state == States.COMPLETED if self.is_completed else None

    @property
    def is_wrong(self) -> Optional[bool]:
        """
        Check if the challenge has been failed. If not, return None. If completed, return False.

        Returns
        -------
        Optional, :py:class:`bool` :
            If the challenge has been failed. If not, return False. If not completed, return None.

        """
        return self.state == States.FAILED if self.is_completed else None

    def begin(self) -> GeneratorReturnType:
        """
        Begins the challenge.

        Raises
        ------
        :py:exc:`AlreadyCompletedError` :
            If the challenge has already been completed. You cannot start a challenge twice,
            you need to create a new one.
        :py:exc:`AlreadyRunningError` :
            If the challenge is already running.
        :py:exc:`TooManyRetriesError` :
            If the number of failures is greater than the number of retries allowed.
            In other words, the challenge has failed.
        :py:exc:`ChallengeCompletionError` :
            If the challenge had a failure. Returns the failure's reason.

        Returns
        -------
        The Captcha object to send to the user.


        .. versionchanged:: 2.1.0

           The return type will now be dynamically acquired and adapt to the given generator.
        """
        if self.state == States.FAILED:
            reason = self.fail_reason or "Failed (No failed reason)"
            raise TooManyRetriesError(reason)
        if self.state == States.FAILURE:
            reason = self.fail_reason or "Failure (No failure reason)"
            raise ChallengeCompletionError(reason)
        if self.state == States.COMPLETED:
            raise AlreadyCompletedError("Challenge already completed")
        if self.state == States.WAITING:
            raise AlreadyRunningError("Challenge is already being ran")

        self._set_state(States.WAITING)
        return self.captcha_object

    def check(
        self, answer: str, *, force_casing: bool = False, remove_spaces: bool = True
    ) -> bool:
        """
        Check an answer.
        This will always add +1 to `attempted_tries` and `failures` if necessary.

        Parameters
        ----------
        answer : :py:class:`str`
            The answer to check against the Captcha's code.
        force_casing : :py:class:`bool`
            If True, the casing must be respected. Defaults to False.
        remove_spaces : :py:class:`bool`
            If True, spaces will be removed when checking the answer. Defaults to True.

        Raises
        ------
        :py:exc:`TooManyRetriesError` :
            If the number of failures is greater than the number of retries allowed.
            We are still adding +1 to the failure even when raising the exception.
        :py:exc:`TypeError` :
            The challenge cannot be edited (State is either not PENDING or not WAITING)

        Return
        ------
        :py:class:`bool` :
            True if the answer is correct, False otherwise.
        """
        if not self._can_be_modified:
            raise TypeError("Challenge cannot be edited")

        self.attempted_tries += 1

        if not self.captcha.check(answer, force_casing=force_casing, remove_spaces=remove_spaces):
            # If wrong
            self.failures += 1
            if self.failures >= self.allowed_retries:
                self._set_state(States.FAILED, FailReason.TOO_MANY_RETRIES)
                raise TooManyRetriesError(self.fail_reason)
            return False
        else:
            # If correct
            self._set_state(States.COMPLETED)
            return True

    def reload(
        self, *, increase_attempted_tries: bool = True, increase_failures: bool = False
    ) -> GeneratorReturnType:
        """
        Reload the Challenge and its code.

        This method will create a new random code + captcha object. It will also increase the
        attempted_tries counter if requested. By defaults, this behavior is executed.

        Parameters
        ----------
        increase_attempted_tries : :py:class:`bool`
            If True, the attempted_tries counter will be increased.

        increase_failures : :py:class:`bool`
            If True, the failures counter will be increased.

        Raises
        ------
        :py:exc:`TypeError` :
            If the challenge cannot be edited or is not already running.

        Returns
        -------
        :py:attr:`discapty.constants.GeneratorReturnType` :
            The Captcha object to send to the user.


        .. versionchanged:: 2.1.0
           The return type will now be dynamically acquired and adapt to the given generator.

        """
        if not self._can_be_modified:
            raise TypeError("Challenge cannot be edited")
        if self.state == States.PENDING:
            raise TypeError("Challenge is not running")

        self.code = random_code()

        if increase_attempted_tries:
            self.attempted_tries += 1
            if increase_failures:
                self.failures += 1

        return self.captcha_object

    def cancel(self) -> None:
        """
        Cancel the challenge.

        Raises
        ------
        :py:exc:`TypeError` :
            If the challenge cannot be edited.
        """
        if not self._can_be_modified:
            raise TypeError("Challenge cannot be edited")

        self._set_state(States.FAILURE, FailReason.CANCELLED)
